/*$Id: parms.c,v 1.1.1.1 2006/11/24 21:15:27 zli Exp $*/
/*

*/

#include "src/ksp/pc/pcimpl.h"          /*I "petscpc.h" I*/
EXTERN_C_BEGIN
#include "psparslib.h"
EXTERN_C_END

/* 
   Private context (data structure) for the  preconditioner.  
*/
typedef struct {
  int                      nloc;
  PARMS_DistMatrix         A;
  PARMS_Vec                rhs, sol;
  PetscScalar              *wk1, *wk2;
  PARMS_PreCon             prec;
  PARMS_PrePar             prepar;
  PARMS_IterPar            ipar;
} PC_PARMS;

static int set_pc_method(char *pc_name, int *method)
{
  int ierr;
  
  ierr = 0;
  if(!strcmp(pc_name, "add_ilu0")) *method = 0;
  else if(!strcmp(pc_name, "add_ilut")) *method = 1;
  else if(!strcmp(pc_name, "add_iluk")) *method = 2;
  else if(!strcmp(pc_name, "add_arms")) *method = 3;
  else if(!strcmp(pc_name, "lsch_ilu0")) *method = 4;
  else if(!strcmp(pc_name, "lsch_ilut")) *method = 5;
  else if(!strcmp(pc_name, "lsch_iluk")) *method = 6;
  else if(!strcmp(pc_name, "lsch_arms")) *method = 7;
  else if(!strcmp(pc_name, "rsch_ilu0")) *method = 8;
  else if(!strcmp(pc_name, "rsch_ilut")) *method = 9;
  else if(!strcmp(pc_name, "rsch_iluk")) *method = 10;
  else if(!strcmp(pc_name, "rsch_arms")) *method = 11;
  else if(!strcmp(pc_name, "sch_sgs")) *method = 12;
  else if(!strcmp(pc_name, "sch_gilu0")) *method = 13;
  else ierr = 1; 
  return ierr;
}


static void Set_PARMS_Params(FILE *fp, PARMS_PrePar prepar,
			     PARMS_IterPar ipar, int *method) 
{
#define BUFLEN 80  
  int i, ierr, num;
  char meth_buf[30];
  char buf[BUFLEN];

  if (fp == 0) {
    ipar->pgfpar[0] = 0.01;
    ipar->pgfpar[1] = 0.01;
    ipar->ipar[0] = 4;
    ipar->ipar[1] = 20;
    ipar->ipar[2] = 0;
    ipar->ipar[3] = 5;
    ipar->ipar[4] = 5;
  
    for (i = 0; i < 7; i++) {
      prepar->droptol[i] = 0.001;
    }
    prepar->mc = 1;
    for (i = 0; i < 7; i++) {
      prepar->lfil[i] = 20;
    }
    prepar->tolind = 0.1;
    *method = 5;
    memcpy(prepar->ipar, ipar->ipar, 18*sizeof(int));
  }
  else {
    for (i = 0; i < 18; i++) {
      ipar->ipar[i] = 0;
    }

    num = 0;
    while(fgets(buf, BUFLEN, fp) != NULL) {
      switch(num) {
      case 0:
      case 1:
      case 2:
      case 3: 
	break;
      case 4:                             /* method */
	sscanf(buf,"%s", meth_buf);
	ierr = set_pc_method(meth_buf, method);
	break;
      case 5:                             /* inner tol */
	sscanf(buf,"%lf",&ipar->pgfpar[0]);
	break;
      case 6:                             /* outer tol */
	sscanf(buf,"%lf",&ipar->pgfpar[1]);
	break;
      case 7:                             /* nlev     */ 
	sscanf(buf,"%d",&ipar->ipar[0]);
	break;
      case 8:                             /* bsize   */
	sscanf(buf,"%d",&ipar->ipar[1]);
	break;
      case 9:                             /* tolind  */
	sscanf(buf,"%lf",&prepar->tolind);
	break;
      case 10:                             /* im (Outer Krylov subs dim) */
	sscanf(buf,"%d",&ipar->ipar[5]);
	break;
      case 11:                             /* max outer gmres steps */
	sscanf(buf,"%d", &ipar->ipar[6]);
	break;
      case 12:                             /* inim (inned Krylov subs dim) */
	sscanf(buf, "%d", &ipar->ipar[3]);
	break;
      case 13:                             /* itsgmr (inner max iterations) */
	sscanf(buf, "%d", &ipar->ipar[4]);
	break;
      case 14:                             /* lfil0 (ilut, iluk, and arms) */
	sscanf(buf, "%d", &prepar->lfil[0]);
	break;
      case 15:                             /* lfil 4 (schur construction) */
	sscanf(buf, "%d", &prepar->lfil[4]);
	break;
      case 16:                             /* lfil 5 (ILUT for Schur) */
	sscanf(buf, "%d", &prepar->lfil[5]);
	break;
      case 17:                             /*droptol (0) -- simliar to lfil0*/
	sscanf(buf, "%lf", &prepar->droptol[0]);
	break;
      case 18:                             /*droptol (4) -- simliar to lfil4*/
	sscanf(buf, "%lf", &prepar->droptol[4]);
	break;
      case 19:                             /*droptol (5) -- simliar to lfil5*/
	sscanf(buf, "%lf", &prepar->droptol[5]);
	break;
      case 20:                             /* multicoloring or not */ 
	sscanf(buf, "%d", &prepar->mc);
	break;
      default:
	break;
      }
      num++;
    }
    if (ierr) {
      fprintf(stderr, "invalid choice for the preconditioner \n");
    }
    memcpy(prepar->ipar, ipar->ipar, 18*sizeof(int));
    
    prepar->droptol[1] = prepar->droptol[2] = prepar->droptol[3] =
      prepar->droptol[0]; 
    prepar->droptol[6] = prepar->droptol[5];
    
    prepar->lfil[1] = prepar->lfil[2] = prepar->lfil[3] =
      prepar->lfil[0];
    prepar->lfil[6] = prepar->lfil[5];
    ipar->in_iters = 0;
  }
}



#undef __FUNCT__  
#define __FUNCT__ "PCSetUp_PARMS"
static int PCSetUp_PARMS(PC pc)
{
  PC_PARMS           *jac = (PC_PARMS*)pc->data;
  PARMS_DistMatrix   A;
  PARMS_Vec          rhs, sol;
  int                *maptmp, *mapptr, npro, rank, low, high;
  int                m, n, lsize, *ib, *jb, *jb1, ncols, pos, i, ierr;
  int                jpos, *cols, j, length, method;
  PetscScalar        *bv, *bv1, *values;
  FILE               *fp;
  
  PetscFunctionBegin;
  MPI_Comm_size(PETSC_COMM_WORLD, &npro);
  MPI_Comm_rank(PETSC_COMM_WORLD, &rank);
  MatGetSize(pc->pmat, &m, &n);
  MatGetOwnershipRange(pc->pmat, &low, &high);
  PetscMalloc((npro+1)*sizeof(int), &mapptr);
  PetscMalloc(n*sizeof(int), &maptmp);
  lsize = high - low;
  jac->nloc = lsize;
  MPI_Allgather(&lsize, 1, MPI_INT, mapptr+1, 1, MPI_INT, PETSC_COMM_WORLD);
  mapptr[0] = 0;
  for (i = 0; i < npro; i++) {
    mapptr[i+1] += mapptr[i];
  }
  for (i = 0; i < n; i++) {
    maptmp[i] = i;
  }

  if (!jac->A) { /* create the matrix the first time through */ 
    parms_MatCreate(&A, "csr");
    parms_GetMap(A, maptmp, mapptr, &n);
    PetscFree(maptmp);
    PetscFree(mapptr);
    PetscMalloc((lsize+1)*sizeof(int), &ib);
    ib[0] = 0;
    length = lsize * 8;
    PetscMalloc(length*sizeof(int), &jb);
    PetscMalloc(length*sizeof(PetscScalar), &bv);

    for (i = low; i < high; i++) {
      pos = ib[i-low];
      MatGetRow(pc->mat, i, &ncols, &cols, &values);
      ib[i-low+1] = ib[i-low] + ncols;
      if (ib[i-low+1] >= length) {
	length += 10;
	PetscMalloc(length*sizeof(int), &jb1);
	memcpy(jb1, jb, ib[i-low]*sizeof(int));
	PetscFree(jb);
	jb = jb1;
	PetscMalloc(length*sizeof(PetscScalar), &bv1);
	memcpy(bv1, bv, ib[i-low]*sizeof(PetscScalar));
	PetscFree(bv);
	bv = bv1;
      }

      memcpy(&jb[pos], cols,  ncols*sizeof(int));
      memcpy(&bv[pos], values,ncols*sizeof(PetscScalar));
      MatRestoreRow(pc->mat, i, &ncols, &cols, &values);
    }

    parms_CopyCsrToDM(A, bv, jb, ib);
    PetscFree(bv);
    PetscFree(jb);
    PetscFree(ib);
    parms_Bdry(A);
    parms_Setup(A);
  }
  if (!jac->rhs) {
    parms_VecCreate(&(jac->rhs));
    parms_CopyComm(A, jac->rhs);
    parms_VecCreate(&(jac->sol));
    parms_CopyComm(A, jac->sol);
    PetscMalloc(lsize*sizeof(PetscScalar), &(jac->wk1));
    PetscMalloc(lsize*sizeof(PetscScalar), &(jac->wk2));
  }
  PetscMalloc(sizeof(*(jac->prepar)), &jac->prepar);
  PetscMalloc(sizeof(*(jac->ipar)),   &jac->ipar);

  fp = fopen("inputs", "r");
  
  Set_PARMS_Params(fp, jac->prepar, jac->ipar, &method);
  if (fp) {
    fclose(fp);
  }

  parms_PrecCreate(A, &jac->prec, method, jac->prepar, jac->ipar);
  jac->A   = A;
  PetscFunctionReturn(0);
}


#undef __FUNCT__  
#define __FUNCT__ "PCApply_PARMS"
static int PCApply_PARMS(PC pc,Vec b,Vec x)
{
  PC_PARMS           *jac = (PC_PARMS*)pc->data;
  int                rank, ierr, i;
  PetscScalar        *bv,*xv;
  PARMS_Vec          pbv,pxv;
  PetscScalar        *sbv,*sxv;
  PARMS_PreCon       prec;
  FILE               *fp;
  char               foo[100];

  PetscFunctionBegin;

  VecGetArray(b, &bv);
  VecGetArray(x, &xv);
  pbv = jac->rhs;
  pxv = jac->sol;
  
  parms_VecAssign(pbv, bv);

  MPI_Comm_rank(PETSC_COMM_WORLD, &rank);

  parms_VecPerm(pbv,   jac->wk1);
  parms_VecAssign(pbv, jac->wk1);
  parms_VecAssign(pxv, jac->wk2);
  prec = jac->prec;

/*   sprintf(foo, "rhs_%d.dat", rank); */
/*   if ((fp = fopen(foo, "w")) == NULL) { */
/*     fprintf(stderr, "cannot open file perm.dat\n"); */
/*     exit(1); */
/*   } */
/*   for (i = 0; i < jac->nloc; i++) { */
/*     fprintf(fp, "%f\n", jac->wk1[i]); */
/*   } */
/*   fclose(fp); */
  prec->prec_operation(jac->A, prec, jac->ipar, pbv, pxv);
/*   sprintf(foo, "sol_%d.dat", rank); */
/*   if ((fp = fopen(foo, "w")) == NULL) { */
/*     fprintf(stderr, "cannot open file perm.dat\n"); */
/*     exit(1); */
/*   } */
/*   for (i = 0; i < jac->nloc; i++) { */
/*     fprintf(fp, "%f\n", jac->wk2[i]); */
/*   } */
/*   fclose(fp); */

  parms_VecInvPerm(pxv, xv);

  VecRestoreArray(b, &bv);
  VecRestoreArray(x, &xv);
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PCDestroy_PARMS"
static int PCDestroy_PARMS(PC pc)
{
  PC_PARMS *jac = (PC_PARMS*)pc->data;
  int      ierr;

  PetscFunctionBegin;
  ierr = PetscFree(jac);CHKERRQ(ierr);
  PetscFunctionReturn(0);
}


/*
    It only gets here if the PARMS type has not been set before the call to 
   ...SetFromOptions() which actually is most of the time
*/
#undef __FUNCT__  
#define __FUNCT__ "PCSetFromOptions_PARMS"
static int PCSetFromOptions_PARMS(PC pc)
{
  return 0;
}

/*MC
     PCPARMS - Allows you to use the matrix element based preconditioners in the LLNL package hypre

   Options Database Keys:
+   -pc_hypre_type - One of pilut, parasails, boomerAMG, euclid
-   Too many others to list, run with -pc_type hypre -pc_hypre_type XXX -help to see options for the XXX
          preconditioner
 
   Level: intermediate

   Notes: Apart from pc_hypre_type (for which there is PCPARMSSetType()),
          the many hypre options can ONLY be set via the options database (e.g. the command line
          or with PetscOptionsSetValue(), there are no functions to set them)

.seealso:  PCCreate(), PCSetType(), PCType (for list of available types), PC,
           PCPARMSSetType()

M*/

EXTERN_C_BEGIN
#undef __FUNCT__  
#define __FUNCT__ "PCCreate_PARMS"
int PCCreate_PARMS(PC pc)
{
  PC_PARMS *jac;
  int       ierr;

  PetscFunctionBegin;
  ierr                     = PetscNew(PC_PARMS,&jac);CHKERRQ(ierr);
  ierr                     = PetscMemzero(jac,sizeof(PC_PARMS));CHKERRQ(ierr);
  pc->data                 = jac;
  jac->A                   = 0;
  jac->rhs                 = 0;
  jac->sol                 = 0;

  pc->ops->setup          = PCSetUp_PARMS;
  pc->ops->apply          = PCApply_PARMS;
  pc->ops->destroy        = PCDestroy_PARMS;

  PetscFunctionReturn(0);
}
EXTERN_C_END
