/*--------------------------------------------------------------------

  parms_OperatorCreate 
  parms_OperatorFree 
  parms_OperatorApply 
  parms_OperatorLsol 
  parms_OperatorInvS 
  parms_OperatorAscend 
  parms_OperatorGetSchurPos
  parms_OperatorGetNnz
  parms_OperatorView 

  Users are NOT encouraged to call those functions directly. 

  $Id: parms_operator.c,v 1.1.1.1 2006/11/24 21:15:30 zli Exp $
  ------------------------------------------------------------------*/
#include "parms_opt_impl.h"

static int parms_Operator_count = 0;

int parms_OperatorFree(parms_Operator *self)
{
  PARMS_VALIDOPTPTR(*self);

  PARMS_OBJSUBREF(*self);
  if (PARMS_OBJGETREF(*self) == 0) {
    return (*self)->o_vptr->dtor((parms_Object *)self);
  }
  return 0;
}

int parms_OperatorCreate(parms_Operator *self)
{
  parms_Operator newOpt;

  PARMS_OBJCREATE(newOpt, (o_dtor)0, (o_view)0);
  newOpt->handler = PARMS_OBJECT_SET_CLASS(PARMS_OPERATOR_CLASS) +
    parms_Operator_count++;
  newOpt->d_vptr->apply  = 0;
  newOpt->d_vptr->lsol   = 0;
  newOpt->d_vptr->invs   = 0;
  newOpt->d_vptr->ascend = 0;
  *self = newOpt;
  return 0;
}

int parms_OperatorView(parms_Operator self, parms_Viewer v)
{
  PARMS_VALIDOPTPTR(self);
  PARMS_VALIDVIEWERPTR(v);
  return self->o_vptr->view((parms_Object)self, v);
}

int parms_OperatorApply(parms_Operator self, parms_Vec y, parms_Vec x) 
{
  PARMS_VALIDOPTPTR(self);
  PARMS_VALIDVECPTR(y);
  PARMS_VALIDVECPTR(x);

  return self->d_vptr->apply(self, y, x);
}

int parms_OperatorLsol(parms_Operator self, FLOAT *y, FLOAT *x)
{
  PARMS_VALIDOPTPTR(self);
  return self->d_vptr->lsol(self, y, x);
}
  
int parms_OperatorInvS(parms_Operator self, FLOAT *y, FLOAT *x)
{
  PARMS_VALIDOPTPTR(self);
  return self->d_vptr->invs(self, y, x);
}

int parms_OperatorAscend(parms_Operator self, FLOAT *y, FLOAT *x)
{
  PARMS_VALIDOPTPTR(self);
  return self->d_vptr->ascend(self, y, x);
}

int parms_OperatorGetSchurPos(parms_Operator self)
{
  PARMS_VALIDOPTPTR(self);
  return self->d_vptr->getssize(self);
}

void parms_OperatorGetNnz(parms_Operator self, int *nnz_mat, int
			  *nnz_pc)
{
  PARMS_VALIDOPTPTR(self);
  self->d_vptr->getnnz(self, nnz_mat, nnz_pc);
}
